#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "monopoly.h"

int receive_packet (void *buf, int maxlen, int player)
{
  int res = -1;

  while (res <= 0) {
    res = recv (mysocket[player], buf, maxlen, 0);
    if (res > 0) {
#ifdef DEBUG
      printf ("  Received %d bytes.\n", res);
      //printf ("  You sent: %s.\n", buf);
#endif
    }
    else if (res == -1) {
      printf ("Didn't getcha...\n");
      printf ("Error %d (%s)\n", errno, strerror (errno));
      printf ("mysocket[%d]=%d.\n", player, mysocket[player]);
      printf ("maxlen=%d.\n", maxlen);
      printf ("Note: EFAULT=%d.\n", EFAULT);
      sleep (1);
    }
    else {
      printf ("res is exactly zero! Please kill me :-) ...\n");
#ifdef DEBUG
      // Die immediately
      printf ("DEBUG enabled... dying.\n");
      exit (4);
#endif
      res = -1;			// Force it to retry
      sleep (5);
    }
  }
  return res;
}

int start_receive ()
{
  int yourlen, res;
  int backlog = 1;
  int opt = 1;
  int mysock;
  int l = sizeof (opt);
  int maxlen = 200;
  struct sockaddr_in my_addr;
  char *buf;

  buf = (char *) malloc (200);
  MYPORT = 2000;

  my_addr.sin_family = AF_INET;
  my_addr.sin_addr.s_addr = htonl (INADDR_ANY);	// My IP Address
  bzero (&(my_addr.sin_zero), 8);	// zeros for fun!
  bzero (buf, maxlen);

  fflush (NULL);
  mysock = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
  setsockopt (mysock, SOL_SOCKET, SO_REUSEADDR, &opt, l);
  res = -1;
  while (res == -1) {
#ifdef DEBUG
    printf ("Listening on %d.\n", MYPORT);
#endif
    my_addr.sin_port = htons (MYPORT);	// must be in network byte order...
    res = bind (mysock, (struct sockaddr *) &my_addr, sizeof (struct sockaddr));
#ifdef DEBUG
    if (res < 0)
      printf ("debug: bind: res=%d, err=%s (%d).\n", res, strerror (errno), errno);
    else
      printf ("Bound to port Okay.\n");
#endif
    if ((res == -1) && (errno == EADDRINUSE))
      MYPORT++;
  }
  yourlen = sizeof (struct sockaddr);
  res = listen (mysock, backlog);
  mysocket[1] = mysock;
  return MYPORT;
}

void open_port (int player)
{
  int res;
  int opt = 1;
  int l = sizeof (opt);

  mysocket[player] = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
  setsockopt (mysocket[player], SOL_SOCKET, SO_REUSEADDR, &opt, l);
  res = -1;
  while (res == -1) {
    res = connect (mysocket[player], (struct sockaddr *) &your_addr[player], sizeof (struct sockaddr));
    if (res == -1) {
      printf ("Error connecting - Errno %d (%s)\n", errno, strerror (errno));
      printf ("  mysocket=%d.\n", mysocket[player]);
      printf ("Connecting to Player %d on port %d\n", player, ntohs (your_addr[player].sin_port));
      sleep (5);
    }
  }
#ifdef DEBUG
  printf ("open_port: Connected to socket %d Okay.\n", mysocket[player]);
#endif

}

void close_port (int player)
{
  close (mysocket[player]);
}

int send_packet (void *buf, int player, int len)
{
  int res;
#ifdef DEBUG
  printf ("Sending to Player %d... (socket %d)\n", player, mysocket[player]);
#endif
  res = send (mysocket[player], buf, len, MSG_DONTWAIT);
  if (res != len)
    printf ("send: res=%d (/%d): %s (%d)\n", res, len, strerror (errno), errno);
  return res;
}

int receive_update ()
{
  int yourlen;
  int mysock;
  int maxlen = 200;
  struct sockaddr_in my_addr;
  char *buf;

  buf = (char *) malloc (200);

#ifdef DEBUG
  printf ("Starting receive_update...\n");
#endif
  my_addr.sin_family = AF_INET;
  // my_addr.sin_port=htons(MYPORT); // must be in network byte order...
  my_addr.sin_addr.s_addr = htonl (INADDR_ANY);	// My IP Address
  bzero (&(my_addr.sin_zero), 8);	// zeros for fun!
  bzero (buf, maxlen);

  mysock = mysocket[1];
  mysocket[1] = accept (mysock, (struct sockaddr *) &your_addr, &yourlen);
#ifdef DEBUG
  printf ("Done Accept()... mysocket[1]=%d\n", mysocket[1]);
#endif
  receive_packet (buf, 12, 1);
  sscanf (buf, "%d %d %d", &MyPlayerNum, &NoOfPlayers, &rules);
#ifdef DEBUG
  printf ("MyPlayerNum=%d, NoOfPlayers=%d.\n", MyPlayerNum, NoOfPlayers);
#endif
  return receive_packet (Names, sizeof (Names), 1);
}

void update_players ()
{
  int player;
  char *buf;

  buf = (char *) malloc (200);

  for (player = 2; player <= NoOfPlayers; player++) {
    printf ("Updating Player %d...\n", player);
    //sleep(3);
    open_port (player);
    sprintf (buf, "%d %d %08d", player, NoOfPlayers, rules);
#ifdef DEBUG
    printf ("SENDING %d bytes\n", strlen (buf));
    printf ("RULES = %d\n", rules);
#endif
    send_packet (buf, player, strlen (buf));
    send_packet (Names, player, sizeof (Names));
  }
  free (buf);
}

void master_wait_on_players ()
{
  char *buf;
  int player = 2;
  int backlog = 1;
  int mysocket, yoursocket;
  struct sockaddr_in my_addr;
  struct sockaddr_in your_addr2;
  int yourlen, res;
  int opt = 1;
  int l = sizeof (opt);
  int port;

  my_addr.sin_family = AF_INET;
  my_addr.sin_port = htons (MYPORT);	// must be in network byte order...
  my_addr.sin_addr.s_addr = htonl (INADDR_ANY);
  bzero (&(my_addr.sin_zero), 8);	// sockaddr is bigger ...
  // ... than sockaddr_in, so fill with blanks

  buf = (char *) malloc (200);
  while (player <= NoOfPlayers) {
    printf ("Waiting on player %d...\n", player);
    mysocket = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
#ifdef DEBUG
    printf ("Opening socket ... port %d, mysocket=%d.\n", MYPORT, mysocket);
#endif
    res = setsockopt (mysocket, SOL_SOCKET, SO_REUSEADDR, &opt, l);
    if (res == -1)
      printf ("SetSockOpt: Error %d, (%s)\n", errno, strerror (errno));
    res = bind (mysocket, (struct sockaddr *) &my_addr, sizeof (struct sockaddr));
    if (res == -1)
      printf ("Bind: ERROR: %s (%d).\n", strerror (errno), errno);
    yourlen = sizeof (struct sockaddr);
    res = listen (mysocket, backlog);

    fflush (NULL);
    yoursocket = accept (mysocket, (struct sockaddr *) &your_addr2, &yourlen);
#ifdef DEBUG
    printf ("Debug: yoursocket=%d.\n", yoursocket);
#endif
    fflush (NULL);
    res = recv (yoursocket, buf, 50, 0);
    fflush (NULL);
    if (res > 0) {
#ifdef DEBUG
      printf ("  Received %d bytes.\n", res);
#endif
      buf[res] = '\0';
#ifdef DEBUG
      printf ("  MWoP: You sent: %s.\n", buf);
#endif
      sscanf (buf, "%d %s", &port, Names[player]);
      // We always talk to our peers on this port:
      your_addr[player] = your_addr2;
      your_addr[player].sin_port = htons (port);
      res = close (yoursocket);
      if (res == -1)
	printf ("ERROR: I Can't close your socket!\n");
#ifdef DEBUG
      printf ("eow - player=%d\n", player);
#endif
      player++;
      res = close (mysocket);
      if (res == -1)
	printf ("close: ERROR: %s (%d)\n", strerror (errno), errno);


    }
    else {
      printf ("Recv: res is %d, Error %d (%s)\n", res, errno, strerror (errno));
      sleep (1);
    }
  }
  printf ("Players are: me [%s] (localhost)", Names[1]);
  for (player = 2; player <= NoOfPlayers; player++) {
    printf (", %s (%s)", Names[player], inet_ntoa (your_addr[player].sin_addr));
  }
  printf (".\n");
  /*
     At this point, we should inform all players:
     a) what their player number is
     b) what all the players are called.
     c) what all players' IP addresses are.
   */
  update_players ();
  free (buf);

}

void send_structured_packet (int pkttype, int index, int val, char *sval)
{
  int p;
  structured_packet *sp;

  if (text == 0)
    gui_update_cash ();
  sp = (structured_packet *) malloc (sizeof (structured_packet));
  sp->pkttype = pkttype;
  sp->index = index;
  sp->val = val;
  strcpy (sp->sval, sval);

#ifdef DEBUG
  printf ("send_structured_packet sending %d[%d]=%d (%s)\n", pkttype, index, val, sval);
#endif
  if (MyPlayerNum == Master) {
    for (p = 2; p <= NoOfPlayers; p++) {
      /* Don't send to the active player, they won't be listening.
         The exception to this, is if they've just become active. */
      if ((p != Player) || (sp->pkttype == PLAYER))
	send_packet (sp, p, sizeof (structured_packet));
    }
  }
  else {
    send_packet (sp, Master, sizeof (structured_packet));
  }
  free (sp);
}
